#!/usr/bin/python2
#-*- coding:utf-8 -*-

import threading
import datetime
import time
import json
from hashlib import sha256
from ecc import CurveFp, Point
from socket import *

PORT = 7780
HOST = "0.0.0.0"
FLAG = "W3C{This is not the real flag}"

# ----------------------------------------------------------------
# SERVICE
# ----------------------------------------------------------------

def main():
    serve()

def serve():
    s = socket(AF_INET, SOCK_STREAM)
    s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    s.bind((HOST, PORT))
    s.listen(1)
    print "[*] Server started on port %d" % PORT

    while True:
        conn, addr = s.accept()
        thread = threading.Thread(None, serve_client, None, (conn,))
        thread.start()

def serve_client(conn):
    try:
        params = read_params()
        send_params(params, conn)
    except Exception as e:
        print str(e)
        conn.sendall("Invalid params, please alert an admin\n")

    t1 = int(time.time())

    while True:
        try:
            data = conn.recv(1024)

            if (int(time.time()) - t1) > 60 * 10:
                conn.sendall("Timeout\n")
                break

            if data == "" or not client(params, data, conn):
                break
        except:
            conn.sendall("Invalid format\n")
            break

    conn.close()

def client(params, data, conn):
    print "[*] Received message %s" % data

    message, S, E = data.strip().split(",")

    s,e = int(S),int(E)

    if check_sig(params, message, s, e):
        conn.sendall("Good signature\n")
        if message == "Gimme the flag!":
            conn.sendall("Congratulations! The flag is %s\n" % FLAG)
            return False
    else:
        conn.sendall("Bad signature\n")
        return False

    return True

def read_params():
    with open("params.json", "rb") as f:
        params = json.loads(f.read())
    return params

def send_params(params, conn):
    out  = "Current params (regenerated every 24 hours):\n"

    for key in ("p", "a", "b", "q", "Ax", "Ay", "Bx", "By"):
        out += "  {} = {}\n".format(key, params[key])

    now = int(time.time())
    creation_delta = now - params['creation_time']
    expiration_delta = params['expiration_time'] - now

    if expiration_delta < 0 or creation_delta < 0:
        raise Exception('Invalid params')

    out += "Params were created {} ago and will expire in {}.\n".format(
        datetime.timedelta(seconds=creation_delta),
        datetime.timedelta(seconds=expiration_delta),
    )
    out += "-" * 65 + "\n"
    conn.sendall(out)

# ----------------------------------------------------------------
# CRYPTO
# ----------------------------------------------------------------


def hash(m):
    HASH_L = 22
    return int(sha256(m).hexdigest()[:HASH_L], 16)

def hash_combine(P,m):
    return hash(str(P.x()) + ':' + str(P.y()) + ':' + m)

def check_sig(params, m, s, e):
    E = CurveFp(params['p'], params['a'], params['b'])

    assert s < params['q']
    assert e < params['q']

    A = Point(E, params['Ax'], params['Ay'])
    B = Point(E, params['Bx'], params['By'])

    R = s*A + e*B

    e_check = hash_combine(R,m)

    return e_check == e

if __name__ == "__main__":
    main()
